/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Matthieu Leclercq
 * Contributor(s):
 */

package org.objectweb.dream.protocol.causality;

import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.ChunkAlreadyExistsException;
import org.objectweb.dream.message.ExtensibleMessage;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.dream.protocol.ArrowChunk;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Outgoing Message stamper.
 */
public class OutMessageStamperImpl extends AbstractComponent
    implements
      Push,
      CausalityTransformerAttributeController
{

  private MessageManager messageManagerItf;
  private MatrixClock    matrixClockItf;
  private Push           outPushItf;

  private String         arrowChunkName;
  private String         causalityChunkName;

  /**
   * @see Push#push(Message, Map)
   */
  public void push(Message message, Map context) throws PushException
  {
    ArrowChunk arrowChunk = (ArrowChunk) message.getChunk(arrowChunkName);

    if (arrowChunk == null)
    {
      logger.log(BasicLevel.ERROR, "Unable to find arrow chunk named "
          + arrowChunkName + ". The message is dropped");
      return;
    }

    CausalityChunk csltChunk = (CausalityChunk) message
        .getChunk(causalityChunkName);
    if (csltChunk == null)
    {
      if (message instanceof ExtensibleMessage)
      {
        csltChunk = (CausalityChunk) messageManagerItf
            .createChunk(CausalityChunk.TYPE);
        try
        {
          ((ExtensibleMessage) message).addChunk(causalityChunkName,
              CausalityChunk.TYPE, csltChunk);
        }
        catch (ChunkAlreadyExistsException e)
        {
          // cannot happend.
        }
      }
      else
      {
        logger
            .log(
                BasicLevel.ERROR,
                "Unable to find causality chunk in the unextensible message. The message is dropped");
        return;
      }
    }
    Object stamp = matrixClockItf.getStamp(arrowChunk.getProcessIdFrom(),
        arrowChunk.getProcessIdTo());
    csltChunk.setCausalityStamp(stamp);
    outPushItf.push(message, context);
  }

  // ---------------------------------------------------------------------------
  // Implementation of CausalityTransformerAttributeController interface
  // ---------------------------------------------------------------------------

  /**
   * @see CausalityTransformerAttributeController#getArrowChunkName()
   */
  public String getArrowChunkName()
  {
    return arrowChunkName;
  }

  /**
   * @see CausalityTransformerAttributeController#setArrowChunkName(String)
   */
  public void setArrowChunkName(String arrowChunkName)
  {
    this.arrowChunkName = arrowChunkName;
  }

  /**
   * @see CausalityTransformerAttributeController#getCausalityChunkName()
   */
  public String getCausalityChunkName()
  {
    return causalityChunkName;
  }

  /**
   * @see CausalityTransformerAttributeController#setCausalityChunkName(String)
   */
  public void setCausalityChunkName(String causalityChunkName)
  {
    this.causalityChunkName = causalityChunkName;
  }

  // ---------------------------------------------------------------------------
  // Implementation of BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{Push.OUT_PUSH_ITF_NAME,
        MatrixClock.MATRIX_CLOCK_ITF_NAME, MessageManager.ITF_NAME};
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(MatrixClock.MATRIX_CLOCK_ITF_NAME))
    {
      matrixClockItf = (MatrixClock) serverItf;
    }
    else if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
  }

}