/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.utobcast;

import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.IOPushException;
import org.objectweb.dream.InterruptedPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.dream.protocol.utobcast.message.UTOBcastChunk;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Implementation of the DAT component. This component handles DAT messages.
 * Upon message reception, it stores the message into the
 * <code>PendingMessages</code> queue, and then it sends a REP message to
 * pbackup. Finally, it increments the sequence number. Note that this component
 * may trigger a backup election if it detects its failure.
 */
public class DATImpl extends AbstractComponent implements Push
{

  // ---------------------------------------------------------------------------
  // Fields of this class
  // ---------------------------------------------------------------------------

  /**
   * The commonly used name to refer to the <code>pendingMessageInItf</code>
   * interface.
   */
  public static final String  PENDING_MESSAGES_IN_ITF_NAME = "pending-messages-in";

  /** The interface to store messages in the <code>PendingMessage</code> queue. */
  protected Push              pendingMessagesInItf;

  /** The interface to send REP messages to pbackup. */
  protected Push              outPushItf;

  /** The interface to retrieve process membership information. */
  protected ProcessMembership processMembershipItf;

  /** The interface to get/set the sequence number. */
  protected SequenceNumber    sequenceNumberItf;

  /** The interface to elect a new backup. */
  protected BackupElection    backupElectionItf;

  /** The interface to manage message lifecycle. */
  protected MessageManager    messageManagerItf;

  final Object                lock                         = new Object();

  // ---------------------------------------------------------------------------
  // Constructor
  // ---------------------------------------------------------------------------

  /**
   * Constructor.
   */
  public DATImpl()
  {
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Push interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.Push#push(org.objectweb.dream.message.Message,
   *      java.util.Map)
   */
  public void push(Message message, Map context) throws PushException
  {
    UTOBcastChunk chunk = (UTOBcastChunk) message
        .getChunk(UTOBcastChunk.DEFAULT_NAME);
    synchronized (lock)
    {
      if (logger.isLoggable(BasicLevel.DEBUG))
      {
        logger.log(BasicLevel.DEBUG, "Received DAT message " + message);
      }
      // Set the sequence number
      chunk.setSequenceNumber(sequenceNumberItf.getSequenceNumber());
      // Add the message to pendingMessages
      if (logger.isLoggable(BasicLevel.DEBUG))
      {
        logger.log(BasicLevel.DEBUG, "Add message " + message
            + " to pending messages");
      }
      messageManagerItf.duplicateMessage(message, false);
      pendingMessagesInItf.push(message, null);
      // Set the new message type
      chunk.setUTOBcastMessageType(UTOBcastChunk.REP);
      // Set the process To (backup)
      try
      {
        chunk.setProcessTo(processMembershipItf.getBackup());
      }
      catch (InterruptedException e)
      {
        // Release resources
        messageManagerItf.deleteMessage(message);
        throw new InterruptedPushException(e);
      }
      // Send REP message to backup
      try
      {
        if (logger.isLoggable(BasicLevel.DEBUG))
        {
          logger.log(BasicLevel.DEBUG, "Send REP message " + message
              + " to backup");
        }
        outPushItf.push(message, null);
      }
      catch (InterruptedPushException e)
      {
        // Release resources
        messageManagerItf.deleteMessage(message);
        throw e;
      }
      catch (IOPushException e)
      {
        // We have detected that backup has crashed
        backupElectionItf.elect();
        return;
      }
      // Increment the sequence number
      if (logger.isLoggable(BasicLevel.DEBUG))
      {
        logger.log(BasicLevel.DEBUG, "Increment sequence number");
      }
      sequenceNumberItf.incrementSequenceNumber();
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
   *      java.lang.Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(PENDING_MESSAGES_IN_ITF_NAME))
    {
      pendingMessagesInItf = (Push) serverItf;
    }
    else if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(SequenceNumber.ITF_NAME))
    {
      sequenceNumberItf = (SequenceNumber) serverItf;
    }
    else if (clientItfName.equals(ProcessMembership.ITF_NAME))
    {
      processMembershipItf = (ProcessMembership) serverItf;
    }
    else if (clientItfName.equals(BackupElection.ITF_NAME))
    {
      backupElectionItf = (BackupElection) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }

  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{PENDING_MESSAGES_IN_ITF_NAME, SequenceNumber.ITF_NAME,
        Push.OUT_PUSH_ITF_NAME, ProcessMembership.ITF_NAME,
        BackupElection.ITF_NAME, MessageManager.ITF_NAME};
  }

}