/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.utobcast;

import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.InterruptedPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.protocol.utobcast.message.UTOBcastChunk;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Implementation of the DeliveredMessageRouter component. This component routes
 * delivered messages on one of its outputs. It has one output for messages that
 * were broadcast by the node itself, and one output for messages broadcast by
 * other nodes.
 */
public class DeliveredMessageRouterImpl extends AbstractComponent
    implements
      Push
{

  // ---------------------------------------------------------------------------
  // Fields of this class
  // ---------------------------------------------------------------------------

  /** The client interfaces. */
  Push                localMessageItf;
  Push                remoteMessageItf;
  ProcessMembership   processMembershipItf;

  short               myProcessId;
  boolean             configured              = false;

  static final String LOCAL_MESSAGE_ITF_NAME  = "local-message";
  static final String REMOTE_MESSAGE_ITF_NAME = "remote-message";

  // ---------------------------------------------------------------------------
  // Constructor
  // ---------------------------------------------------------------------------

  /**
   * Constructor.
   */
  public DeliveredMessageRouterImpl()
  {
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Push interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.Push#push(org.objectweb.dream.message.Message,
   *      java.util.Map)
   */
  public void push(Message message, Map context) throws PushException
  {
    if (!configured)
    {
      try
      {
        configure();
      }
      catch (InterruptedException e)
      {
        throw new InterruptedPushException(e);
      }
    }
    UTOBcastChunk chunk = (UTOBcastChunk) message
        .getChunk(UTOBcastChunk.DEFAULT_NAME);

    if (chunk.getProcessFrom().getId() == myProcessId)
    {
      if (logger.isLoggable(BasicLevel.DEBUG))
      {
        logger.log(BasicLevel.DEBUG, "Deliver a local message " + message);
      }
      localMessageItf.push(message, context);
    }
    else
    {
      if (logger.isLoggable(BasicLevel.DEBUG))
      {
        logger.log(BasicLevel.DEBUG, "Deliver a remote message " + message);
      }
      remoteMessageItf.push(message, context);
    }
  }

  // ---------------------------------------------------------------------------
  // Utility methods.
  // ---------------------------------------------------------------------------

  void configure() throws InterruptedException
  {
    myProcessId = processMembershipItf.getMyself().getId();
    configured = true;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
   *      java.lang.Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(LOCAL_MESSAGE_ITF_NAME))
    {
      localMessageItf = (Push) serverItf;
    }
    else if (clientItfName.equals(REMOTE_MESSAGE_ITF_NAME))
    {
      remoteMessageItf = (Push) serverItf;
    }
    else if (clientItfName.equals(ProcessMembership.ITF_NAME))
    {
      processMembershipItf = (ProcessMembership) serverItf;
    }

  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{LOCAL_MESSAGE_ITF_NAME, REMOTE_MESSAGE_ITF_NAME,
        ProcessMembership.ITF_NAME};
  }

}