/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.protocol.utobcast;

import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.IOPushException;
import org.objectweb.dream.InterruptedPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.dream.protocol.Process;
import org.objectweb.dream.protocol.utobcast.message.UTOBcastChunk;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Implementation of the REP component. This component handles REP messages.
 * Upon message reception, it stores the message into the
 * <code>PendingMessages</code> queue. Then it sends UTO messages all the
 * processes but pleader, and an ACK message to pleader. Note that in case this
 * component detects a failure of pleader, it becomes leader, and trigger a
 * backup election.
 */
public class REPImpl extends AbstractComponent implements Push
{

  // ---------------------------------------------------------------------------
  // Fields of this class
  // ---------------------------------------------------------------------------

  /**
   * The commonly used name to refer to the <code>pendingMessageInItf</code>
   * interface.
   */
  public static final String  PENDING_MESSAGES_IN_ITF_NAME = "pending-messages-in";

  /** The interface to store messages in the <code>PendingMessage</code> queue. */
  protected Push              pendingMessagesInItf;

  /** The interface to send UTO and ACK messages. */
  protected Push              outPushItf;

  /** The interface to retrieve process membership information. */
  protected ProcessMembership processMembershipItf;

  /** The interface to elect a new backup. */
  protected BackupElection    backupElectionItf;

  /** The interface to manage message lifecycle. */
  protected MessageManager    messageManagerItf;

  Object                      lock                         = new Object();

  // ---------------------------------------------------------------------------
  // Constructor
  // ---------------------------------------------------------------------------

  /**
   * Constructor.
   */
  public REPImpl()
  {
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Push interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.Push#push(org.objectweb.dream.message.Message,
   *      java.util.Map)
   */
  public void push(Message message, Map context) throws PushException
  {
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "Received REP message " + message);
    }
    ///////////////////////////////////////////////////////////
    // Add message to pendingMessages
    ///////////////////////////////////////////////////////////
    if (logger.isLoggable(BasicLevel.DEBUG))
    {
      logger.log(BasicLevel.DEBUG, "Add REP message " + message
          + " to PendingMessages");
    }
    messageManagerItf.duplicateMessage(message, false);
    pendingMessagesInItf.push(message, null);

    ///////////////////////////////////////////////////////////
    // Send UTO messages to all processes but leader and backup
    ///////////////////////////////////////////////////////////
    Process[] processes = processMembershipItf.getOtherProcesses();
    UTOBcastChunk chunk = (UTOBcastChunk) message
        .getChunk(UTOBcastChunk.DEFAULT_NAME);
    chunk.setUTOBcastMessageType(UTOBcastChunk.UTO);
    for (int i = 0; i < processes.length; i++)
    {
      // We do not clone messages. Be carefull, only works if messages are
      // sent using this thread.
      // Set the receiver process
      chunk.setProcessTo(processes[i]);
      if (logger.isLoggable(BasicLevel.DEBUG))
      {
        logger.log(BasicLevel.DEBUG, "Send UTO message " + message
            + " to process " + processes[i]);
      }
      try
      {
        messageManagerItf.duplicateMessage(message, false);
        outPushItf.push(message, null);
      }
      catch (InterruptedPushException e)
      {
        // Release resources: should delete duplicated messages
        throw e;
      }
      catch (IOPushException e)
      {
        // The process processes[i] has crashed.
        processMembershipItf.removeProcess(processes[i]);
      }
    }

    ///////////////////////////////////////////////////////////
    //  Send ACK message to leader
    ///////////////////////////////////////////////////////////
    chunk.setUTOBcastMessageType(UTOBcastChunk.ACK);

    try
    {
      Process leader = processMembershipItf.getLeader();
      if (logger.isLoggable(BasicLevel.DEBUG))
      {
        logger.log(BasicLevel.DEBUG, "Send ACK message " + message
            + " to process " + leader);
      }
      chunk.setProcessTo(leader);
    }
    catch (InterruptedException e1)
    {
      throw new InterruptedPushException(e1);
    }
    try
    {
      // We do not duplicate the message
      outPushItf.push(message, null);
    }
    catch (IOPushException e)
    {
      // We detect that the leader has crashed
      // TODO
      e.printStackTrace();
    }

  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
   *      java.lang.Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(PENDING_MESSAGES_IN_ITF_NAME))
    {
      pendingMessagesInItf = (Push) serverItf;
    }
    else if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(ProcessMembership.ITF_NAME))
    {
      processMembershipItf = (ProcessMembership) serverItf;
    }
    else if (clientItfName.equals(BackupElection.ITF_NAME))
    {
      backupElectionItf = (BackupElection) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }

  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{PENDING_MESSAGES_IN_ITF_NAME, Push.OUT_PUSH_ITF_NAME,
        ProcessMembership.ITF_NAME, BackupElection.ITF_NAME,
        MessageManager.ITF_NAME};
  }

}