/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact : dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s):
 */

package org.objectweb.dream.pushwithreturn;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.objectweb.dream.AbstractComponent;
import org.objectweb.dream.InterruptedPushException;
import org.objectweb.dream.Push;
import org.objectweb.dream.Push1;
import org.objectweb.dream.PushException;
import org.objectweb.dream.PushWithReturn;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.dream.pool.ObjectPool;
import org.objectweb.dream.pool.Recyclable;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This component handles calls to the
 * {@link PushWithReturn#pushWithReturn(Message, Map)}method (i.e. outgoing
 * messages). It has one {@link PushWithReturn}input, one {@link Push}output,
 * and one {@link Push1}input.
 * <p>
 * It receives calls to the <code>pushWithReturn</code> method. It generates a
 * key for the message, then it pushes the message on its output, and waits for
 * a message with the same key to arrive on its <code>Push1</code> input. This
 * message is the returned message of the <code>pushWithReturn</code> method.
 * <p>
 * <i>Note: </i> if a message with an "All" key is received on the input, it is
 * used as a return message of all the <code>pushWithReturn</code> method
 * calls currently waiting for a return message.
 */
public class SynchronizerImpl extends AbstractComponent
    implements
      PushWithReturn,
      Push1,
      SynchronizerAttributeController
{

  /** An hash map containing waiting keys. */
  HashMap        waitingKeysMap = new HashMap();

  // Client interfaces
  /** The client interface used to access the message manager. */
  MessageManager messageManagerItf;

  /** The client interface used to access the WaitingKey pool. */
  ObjectPool     waitingKeyPoolItf;

  /** The KeyGenerator client interface. */
  KeyGenerator   keyGeneratorItf;

  /** The Push output interface. */
  Push           outPushItf;

  boolean        mustClone;

  // ---------------------------------------------------------------------------
  // Implementation of the PushWithReturn interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.PushWithReturn#pushWithReturn(org.objectweb.dream.message.Message,
   *      java.util.Map)
   */
  public Message pushWithReturn(Message message, Map context)
      throws PushException
  {
    return doPush(message);
  }

  /**
   * @see Push#push(Message, Map)
   */
  public void push(Message message, Map context) throws PushException
  {
    doPush(message);
  }

  // ---------------------------------------------------------------------------
  // Implementation of the Push1 interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.Push1#push1(org.objectweb.dream.message.Message,
   *      java.util.Map)
   */
  public void push1(Message message, Map context) throws PushException
  {
    Key key = null;
    try
    {
      // Generate or retrieve a key for the message
      key = keyGeneratorItf.generateKey(message);
    }
    catch (Exception e)
    {
      logger.log(BasicLevel.ERROR, e);
      throw new PushException(e);
    }

    if (key.isAll())
    {
      synchronized (waitingKeysMap)
      {
        Iterator iter = waitingKeysMap.keySet().iterator();
        while (iter.hasNext())
        {
          Object keyTemp = iter.next();
          WaitingKey waitingKey = (WaitingKey) waitingKeysMap.get(keyTemp);
          waitingKey.setReturnMessage(message);
          synchronized (waitingKey)
          {
            waitingKey.setCanPass(true);
            waitingKey.notifyAll();
          }
        }
      }
    }
    else
    {
      WaitingKey waitingKey = (WaitingKey) waitingKeysMap.get(key);
      if (waitingKey != null)
      {
        waitingKey.setReturnMessage(message);
        synchronized (waitingKey)
        {
          waitingKey.setCanPass(true);
          waitingKey.notifyAll();
        }
      }
      else
      {
        messageManagerItf.deleteMessage(message);
      }
    }
  }

  // ---------------------------------------------------------------------------
  // Utility methods.
  // ---------------------------------------------------------------------------

  private Message doPush(Message message) throws PushException
  {
    Key key = null;
    try
    {
      // Generate a key for the message
      key = keyGeneratorItf.generateKey(message);
    }
    catch (Exception e)
    {
      logger.log(BasicLevel.ERROR, e);
      throw new PushException(e);
    }

    //Add the key to the HashMap
    WaitingKey waitingKey;
    synchronized (waitingKeysMap)
    {
      waitingKey = (WaitingKey) waitingKeysMap.get(key);
      if (waitingKey == null)
      {
        waitingKey = (WaitingKey) waitingKeyPoolItf.newInstance();
        waitingKeysMap.put(key, waitingKey);
      }
      synchronized (waitingKey)
      {
        waitingKey.incrementNbWaitings();
      }
    }

    // Push the message on the output
    outPushItf.push(message, null);

    // Wait for the return message
    Message returnMessage;
    synchronized (waitingKey)
    {
      if (!waitingKey.canPass())
      {
        try
        {
          waitingKey.wait();
        }
        catch (InterruptedException e)
        {
          throw new InterruptedPushException(e);
        }
      }
      returnMessage = waitingKey.getReturnMessage();
      waitingKey.decrementNbWaitings();
      if (waitingKey.getNbWaitings() > 0)
      {
        // There are other calls waiting
        return messageManagerItf.duplicateMessage(returnMessage, mustClone);
      }
    }
    synchronized (waitingKeysMap)
    {
      synchronized (waitingKey)
      {
        if (waitingKey.getNbWaitings() <= 0)
        {
          waitingKeyPoolItf.recycleInstance((Recyclable) waitingKeysMap
              .remove(key));
        }
      }
    }
    return returnMessage;
  }

  //---------------------------------------------------------------------------
  // Implementation of the SynchronizerAttributeController
  // interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.pushwithreturn.SynchronizerAttributeController#setMustClone(boolean)
   */
  public void setMustClone(boolean mustClone)
  {
    this.mustClone = mustClone;
  }

  /**
   * @see org.objectweb.dream.pushwithreturn.SynchronizerAttributeController#getMustClone()
   */
  public boolean getMustClone()
  {
    return mustClone;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(java.lang.String,
   *      java.lang.Object)
   */
  public void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(Push.OUT_PUSH_ITF_NAME))
    {
      outPushItf = (Push) serverItf;
    }
    else if (clientItfName.equals(KeyGenerator.ITF_NAME))
    {
      keyGeneratorItf = (KeyGenerator) serverItf;
    }
    else if (clientItfName.equals(ObjectPool.ITF_NAME))
    {
      waitingKeyPoolItf = (ObjectPool) serverItf;
    }
    else if (clientItfName.equals(MessageManager.ITF_NAME))
    {
      messageManagerItf = (MessageManager) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{KeyGenerator.ITF_NAME, Push.OUT_PUSH_ITF_NAME,
        ObjectPool.ITF_NAME, MessageManager.ITF_NAME};
  }

  // ---------------------------------------------------------------------------
  // Inner class.
  // ---------------------------------------------------------------------------

  /**
   * This class represents a waiting key.
   */
  public static class WaitingKey implements Recyclable
  {
    /** The return message. */
    private Message returnMessage;

    /** The number of waiting calls. */
    private int     nbWaitings = 0;

    private boolean canPass    = false;

    /**
     * Returns the return message.
     * 
     * @return Returns the message.
     */
    public Message getReturnMessage()
    {
      return returnMessage;
    }

    /**
     * Sets the return message.
     * 
     * @param returnMessage the return message to set.
     */
    public void setReturnMessage(Message returnMessage)
    {
      this.returnMessage = returnMessage;
    }

    /**
     * Increments the number of waiting calls.
     */
    public void incrementNbWaitings()
    {
      nbWaitings++;
    }

    /**
     * Decrements the number of waiting calls.
     */
    public void decrementNbWaitings()
    {
      nbWaitings--;
    }

    /**
     * Returns the number of waiting calls.
     * 
     * @return the number of waiting calls.
     */
    public int getNbWaitings()
    {
      return nbWaitings;
    }

    /**
     * Sets canPass to the given value.
     * 
     * @param canPass the value to which canPass must be set.
     */
    public void setCanPass(boolean canPass)
    {
      this.canPass = canPass;
    }

    /**
     * Returns the value of canPass.
     * 
     * @return the value of canPass.
     */
    public boolean canPass()
    {
      return this.canPass;
    }

    // ---------------------------------------------------------------------------
    // Implementation of the Recyclable interface.
    // ---------------------------------------------------------------------------

    /**
     * @see org.objectweb.dream.pool.Recyclable#recycle()
     */
    public void recycle()
    {
      this.returnMessage = null;
      this.nbWaitings = 0;
      this.canPass = false;
    }
  }

}