/**
 * Dream
 * Copyright (C) 2003-2004 INRIA Rhone-Alpes
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: dream@objectweb.org
 *
 * Initial developer(s): Vivien Quema
 * Contributor(s): 
 */

package org.objectweb.dream.queue;

import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.manager.MessageManager;
import org.objectweb.dream.pool.ObjectPool;
import org.objectweb.dream.pool.Recyclable;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;

/**
 * Implementation of the <code>Buffer</code> interface using a sorted list.
 * This buffer sorts messages according to a sequence number. For that purpose
 * all messages must contain a {@link SequenceNumberChunk}chunk. The name under
 * which this chunk is registered is specified using the
 * {@link QueueSortedAttributeController}.
 * <p>
 * This buffer guarantees that if message <i>m1 </i> (with sequence number <i>sn
 * </i>) is removed from the buffer, then next message to be removed will have
 * sequence number <i>sn+1 </i>. As a consequence, the number of available
 * message may be different from the number of stored messages.
 * <p>
 * <p>
 * <i>Note: </i>This buffer DOES NOT ALLOW overflow policies that drop messages.
 * <i>Indicators </i> on available space, available messages, and stored
 * messages are expressed as number of messages.
 * 
 * @see Buffer
 * @see AbstractBufferImpl @
 */
public class BufferAscendingSequenceNumberImpl extends AbstractBufferImpl
    implements
      QueueSortedAttributeController
{
  /**
   * The name under which the chunk containing the sequence number is
   * registered.
   */
  protected String     sortingChunkName;

  /** The first element stored in this buffer. */
  protected Element    first          = null;

  /** The last element stored in this buffer. */
  protected Element    last           = null;

  /** The client interface used to access the pool of {@link Element}instances. */
  protected ObjectPool objectPoolItf;

  /** The last sequence number in sequence. */
  protected long       lastInSequence = -1;

  // ---------------------------------------------------------------------------
  // Implemented methods.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.queue.AbstractBufferImpl#hasAvailableMessage()
   */
  protected boolean hasAvailableMessage()
  {
    return (availableMessagesIndicator > 0);
  }

  /**
   * @see org.objectweb.dream.queue.AbstractBufferImpl#canAdd(org.objectweb.dream.message.Message)
   */
  protected boolean canAdd(Message message)
  {
    if (maxCapacity > 0 && (maxCapacity - storedMessagesIndicator <= 0))
    {
      return false;
    }
    return true;
  }

  // ---------------------------------------------------------------------------
  // Implemented methods for the Buffer interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.queue.AbstractBufferImpl#doAdd(org.objectweb.dream.message.Message)
   */
  protected void doAdd(Message message)
  {
    long sequenceNumber = ((SequenceNumberChunk) message
        .getChunk(sortingChunkName)).getSequenceNumber();
    if (sequenceNumber <= lastInSequence)
    {
      // The message has already been stored or delivered.
      messageManagerItf.deleteMessage(message);
      return;
    }
    Element current = last;
    while (current != null && sequenceNumber < current.sequenceNumber)
    {
      current = current.previous;
    }

    if (current != null && sequenceNumber == current.sequenceNumber)
    {
      // The message is already stored in the buffer.
      messageManagerItf.deleteMessage(message);
      return;
    }
    Element element = (Element) objectPoolItf.newInstance();
    element.msg = message;
    element.sequenceNumber = sequenceNumber;

    if (current == null)
    {
      // insert at the beginning
      if (first != null)
      {
        // list not empty
        element.next = first;
        first.previous = element;
      }
      else
      {
        // list is empty, first = last
        last = element;
      }
      first = element;
    }
    else
    {
      // insert after current
      element.previous = current;
      if (current.next != null)
      {
        // new element is not the new last
        element.next = current.next;
        element.next.previous = element;
      }
      else
      {
        // insert at the end
        last = element;
      }
      element.previous.next = element;

    }

    // update last in sequence
    //    logger.log(BasicLevel.INFO, "inserted element " +
    // element.sequenceNumber);
    current = element;
    int counter = 0;
    while (current != null && current.sequenceNumber == lastInSequence + 1)
    {
      counter++;
      lastInSequence++;
      current = current.next;
    }
    if (counter > 0)
    {
      incrementAvailableMessagesIndicator(counter);
    }
    incrementStoredMessagesIndicator(1);
  }

  /**
   * @see org.objectweb.dream.queue.AbstractBufferImpl#doRemove()
   */
  protected Message doRemove()
  {
    incrementAvailableMessagesIndicator(-1);
    incrementStoredMessagesIndicator(-1);
    Element newFirst = first.next;
    if (newFirst != null)
    {
      newFirst.previous = null;
    }
    else
    {
      last = null;
    }
    Message msg = first.msg;
    objectPoolItf.recycleInstance(first);
    first = newFirst;
    return msg;
  }

  /**
   * @see org.objectweb.dream.queue.AbstractBufferImpl#doGet()
   */
  protected Message doGet()
  {
    return first.msg;
  }

  // ---------------------------------------------------------------------------
  // Implementation of the QueueSortedAttributeController interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.dream.queue.QueueSortedAttributeController#getSortingChunkName()
   */
  public String getSortingChunkName()
  {
    return sortingChunkName;
  }

  /**
   * @see QueueSortedAttributeController#setSortingChunkName(String)
   */
  public void setSortingChunkName(String name)
  {
    this.sortingChunkName = name;
  }

  // ---------------------------------------------------------------------------
  // Inner class.
  // ---------------------------------------------------------------------------

  /**
   * This class represents elements of a linked list. Each element contains a
   * reference to its predecessor (<code>previous</code>), its successor (
   * <code>next</code>), a message, and a sequence number.
   */
  public static class Element implements Recyclable
  {
    Element previous;
    Element next;
    Message msg;
    long    sequenceNumber;

    // ---------------------------------------------------------------------------
    // Implementation of the Recyclable interface.
    // ---------------------------------------------------------------------------

    /**
     * @see Recyclable#recycle()
     */
    public void recycle()
    {
      previous = null;
      next = null;
      msg = null;
      sequenceNumber = -1;
    }
  }

  // ---------------------------------------------------------------------------
  // Implementation of the BindingController interface.
  // ---------------------------------------------------------------------------

  /**
   * @see org.objectweb.fractal.api.control.BindingController#bindFc(String,
   *      Object)
   */
  public synchronized void bindFc(String clientItfName, Object serverItf)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException
  {
    super.bindFc(clientItfName, serverItf);
    if (clientItfName.equals(ObjectPool.ITF_NAME))
    {
      objectPoolItf = (ObjectPool) serverItf;
    }
  }

  /**
   * @see org.objectweb.fractal.api.control.BindingController#listFc()
   */
  public String[] listFc()
  {
    return new String[]{MessageManager.ITF_NAME, ObjectPool.ITF_NAME};
  }
}